//
//  cpp.cpp
//
//  Test bed for algorithms to compute the Closest Pair of Points
//  Created by Martin Hitz on April 13, 2022.
//  Modified by Martin Hitz on Feb 14, 2025
//  Copyright © 2022-2025 Martin Hitz. All rights reserved.
//

// #define SIGMA  // ND-test with varying sigma at fixed n

#include <iostream>
#include <random>
#include <cmath>
#include <cstdlib> 		// qsort
#include <time.h> 		// used in Log
#include <map>			// used in PSSolver
#include <assert.h> 	// used here and there

#define DIM(array) (sizeof(array) / sizeof(array[0]))

class Log { // Time and steps measurement
    long calls;
    int rec, maxRec;
    long startTime, totalTime;
    double totalSqrTime;
    std::string method;

    static Log* methods[100];
    static int lastMethod;
    static bool header;
    constexpr static const auto delim = " ";


    static long getCpuTime( ) {
        struct timespec now;
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &now);
        return now.tv_sec*1000000000 + now.tv_nsec; // Nanoseconds
    }

public:
    Log (std::string methodName) {
        methods[++lastMethod] = this;
        method = methodName;
        start();
    }

    void start () {
        maxRec = rec = 0;
        calls = 0;
        totalTime = totalSqrTime = 0;
    }

    long elapsedTime () const { return totalTime; } 				// µs
    long meanElapsedTime (int n) const { return totalTime / n; } 	// µs
    long varElapsedTime (int n) const { 							// µs2
        auto mT = meanElapsedTime(n);
        return totalSqrTime / n - mT * mT;
    }

    long getCalls () const { return calls; }
    auto getRec () const { return rec; }

    void enter () {
        if (rec++ == 0)
            startTime = getCpuTime();
        calls++;
        if (rec > maxRec)
            maxRec = rec;
    }

    void exit () {
        if (--rec == 0) {
            auto eTime = (getCpuTime() - startTime) / 1000L; // µs
            totalTime += eTime;
            totalSqrTime += eTime*eTime;
        }
    }

    void print (int n, int times = 1) const {
		std::cout << method << delim << n << delim << getCalls() << delim << (maxRec-1) << delim << meanElapsedTime(times) << delim << varElapsedTime(times) << std::endl;
    }

    static void startAll () {
        for (int i=0; i<=lastMethod; i++)
            methods[i]->start();
    }

    static Log* make (std::string id) {
        for (int i=0; i <= lastMethod; i++)
            if (id.compare(methods[i]->method) == 0)
                return methods[i];
        return new Log(id);
    }
    
    static void printAll (int n, int times = 1) {
        if (!header) {
            std::cout << "Method" << delim << "N" << delim << "Calls" << delim << "Recursions" << delim << "Mean(Time)" << delim << "Var(Time)" << std::endl;
            header = true;
        }
        for (int i=0; i<=lastMethod; i++)
            if (methods[i]->getCalls() > 0)
                methods[i]->print(n, times);
    }
};

int Log::lastMethod = -1;
bool Log::header = false;
Log* Log::methods[100];

class Point {
    double x, y;

public:
    Point (double x=0, double y=0) : x(x), y(y) {}
    
    double getX () const { return x; }
    double getY () const { return y; }
};

struct PointPair {
    const Point * p1, * p2;
};

struct PointList {
    const PointList* next;
    const Point* p;

    PointList (const Point& pt, const PointList* next) : p(&pt), next(next) {}
    
     ~PointList () {
        if (next != NULL)
            delete next;
    }
};

struct MarkedPointList: public PointList {
    bool visited; // Marks 1st point in list: true, if list has already been considered. Unused for other points in list.
    
    MarkedPointList (const Point& pt, const PointList* next) : PointList(pt, next), visited(false) {}
};

class Grid { // Simulation of a large two-dimensional grid using a hash table with O(n) entries
    class HashBucket {
        const int i, j;      // Grid coordinates
        MarkedPointList* pl; // Point list for cell i/j
        HashBucket* next;    // Collision chain -> leads to other i/j-values
    
    public:
        HashBucket (int i, int j, const Point& p, HashBucket * next = NULL) : i(i), j(j), next(next), pl(new MarkedPointList(p, NULL)) {}
        
        ~HashBucket () {
            if (pl != NULL)
                delete pl;
            if (next != NULL)
                delete next;
        }

        int add (int i, int j, const Point& p) {
            int count = 0; // Number of steps
            for (HashBucket * b = this; b != NULL; b = b->next, count++)
                if (b->i == i && b->j == j) { // Found grid cell (i,j)
                    b->pl = new MarkedPointList(p, b->pl);
                    return count;
                }
            // Collision chain exhausted
            next = new HashBucket(i, j, p, next);
            return -count; // Collision occurred if value < 0
        }

        MarkedPointList* find (int i, int j) const {
            for (const HashBucket * b = this; b != NULL; b = b->next)
                if (b->i == i && b->j == j)  // Found grid cell (i,j)
                    return b->pl;
            return NULL; // Grid cell (i,j) not found
        }
	};

    HashBucket ** cells;
    unsigned size, size1, gridSize, n;
    int diffBits;
    int collisions;

    unsigned hash (unsigned i, unsigned j) const {
        unsigned k;
        if (diffBits > 0) { // index > size
            k = ((i * 3) % size);
            k ^= ((j * 11) % size);
        } else {
            k = (i * 3) << -diffBits;
            k ^= (j * 11);
        }
        return k & size1;
    }

public:
    Grid (int g, int nPoints) {
        n = nPoints;
        unsigned sizeBits = int(log2(3*n)) + 1;
        size = (1 << sizeBits);
        size1 = size - 1; // All LLLL...LLL
        gridSize = g;     // Max. index i/j -> g-1
        unsigned indexBits = int(log2(g-1))+1;
        diffBits = indexBits - sizeBits;
    
        cells = new HashBucket*[size];
        collisions = 0;
        for (int i=0; i<size; i++)
            cells[i] = NULL;
    }
    
    ~Grid () {
        for (int i=0; i<size; i++)
            if (cells[i] != NULL)
                delete cells[i];
        delete [] cells;
    }

    int insert (int i, int j, const Point& p) {
        int count = 0;
        int h = hash(i, j);
        if (cells[h] == NULL)
            cells[h] = new HashBucket(i, j, p);
        else {
            count = cells[h]->add(i, j, p);
            if (count < 0) { // collision
                collisions++;
                count = -count;
            }
        }
        return count;
    }

    MarkedPointList* get (int i, int j) const {
        int h = hash(i, j);
        return cells[h] == NULL ? NULL : cells[h]->find(i, j);
    }
    
    int getCollisions () const { return collisions; }
};

class PointGenerator { // Random generators
public:
    virtual Point next () = 0;
    virtual std::string id () const = 0;
};

class UDGenerator: public PointGenerator { // Generator for uniform distributions
    const double xMin, xMax, yMin, yMax;

public:
    UDGenerator (double xMin, double xMax, double yMin, double yMax) :
        xMin(xMin), xMax(xMax), yMin(yMin), yMax(yMax) {
    }

    Point next () {
        return Point(nextDouble() * (xMax - xMin) + xMin,
                     nextDouble() * (yMax - yMin) + yMin);
    }
    
    std::string id () const {
        char id[300];
        sprintf(id, "UD %4.2f %4.2f %4.2f %4.2f", xMin, xMax, yMin, yMax);
        return std::string(id);
    }

private:
    double nextDouble () const { return double(rand()) / double(RAND_MAX); } // Assuming rand() is uniformly distributed from 0 to RAND_MAX
};

class NDGenerator: public PointGenerator { // Generator for truncated normal distributions
    const double xMin, xMax, yMin, yMax;
    const double mx, my;
    const double sigma;
    std::default_random_engine generator;
    std::normal_distribution<double> distribution;

public:
    NDGenerator (double xMin, double xMax, double yMin, double yMax, double s=0.2) :
        xMin(xMin), xMax(xMax), yMin(yMin), yMax(yMax),
        mx((xMax + xMin) / 2), my((yMax + yMin) / 2),
        sigma(s), distribution(0.0, sigma) {}

    Point next () {
        while (true) {
            const double x = distribution(generator) + mx;
            const double y = distribution(generator) + my;
            if (x >= xMin && x <= xMax && y >= yMin && y <= yMax)
                return Point(x, y);
        }
    }
    
    std::string id () const {
        char id[300];
        sprintf(id, "ND %4.2f/%4.2f %4.2f (%4.2f %4.2f %4.2f %4.2f)", mx, my, sigma, xMin, xMax, yMin, yMax);
        return std::string(id);
    }
};

const Point* makePoints (int n, PointGenerator& r) {
    Point* points = new Point[n];
    for (int i=0; i<n; i++) {
        points[i] = r.next();
    }
    return points;
}

Point* clone (const Point* ori, int n) {
    Point* c = new Point[n];
    for (int i = 0; i < n; i++)
        c[i] = ori[i];
    return c;
}

int comparePointsX (const void* p1, const void* p2) {
    double x1 = ((const Point *) p1)->getX();
    double x2 = ((const Point *) p2)->getX();
    
    return x1 < x2 ? -1 : x1 > x2 ? +1 : 0;
}

int comparePointsY (const void* p1, const void* p2) {
    double y1 = ((const Point *) p1)->getY();
    double y2 = ((const Point *) p2)->getY();
    
    return y1 < y2 ? -1 : y1 > y2 ? +1 : 0;
}

int comparePtrPointsX (const void* p1, const void* p2) {
    double x1 = (*((const Point **) p1))->getX();
    double x2 = (*((const Point **) p2))->getX();
    
    return x1 < x2 ? -1 : x1 > x2 ? +1 : 0;
}

int comparePtrPointsY (const void* p1, const void* p2) {
    double y1 = (*((const Point **) p1))->getY();
    double y2 = (*((const Point **) p2))->getY();
    
    return y1 < y2 ? -1 : y1 > y2 ? +1 : 0;
}

bool isSortedX (Point * points, const int n) {
    for (int i = 1; i < n; i++)
        if (points[i-1].getX() > points[i].getX())
            return false;
    return true;
}

void bsort (Point * points, const int n) { // bucket-sorts array points by x
   const static int nEl = 5; // Nominal bucket size

   class Bucket {
          int size = int(nEl * 2.5); // Heurstic
          Point * b; // Array of Points
          int n;
       
      public:
       Bucket () {
           b = new Point[size];
           n = 0;
       }
       
          void insert (const Point & p) {
              if (n >= size) {
                  size = size + nEl;
                  Point * bNew = new Point[size]; // Array of Points
                  for (int i = 0; i < n; i++)
                      bNew[i] = b[i];
                  delete [] b;
                  b = bNew;
              }
              b[n++] = p;
           }
           
           void sort () { qsort(b, n, sizeof(Point), comparePointsX); }
           
           Point * transfer (Point * array) const {
               for (int i = 0; i < n; i++)
                   *(array++) = b[i];
               return array; // next to be transferred
           }
       
           ~Bucket () { delete [] b; }
       };
       
       int bucketNumber = n / nEl + 1;
       Bucket * buckets = new Bucket[bucketNumber]; // Array of buckets of Point pointers

       for (int i = 0; i < n; i++)  //Filling Buckets
           buckets[int((points[i].getX() + 0.5) * bucketNumber)].insert(points[i]);

       for (int i = 0; i < bucketNumber; i++)
           buckets[i].sort();
       
       for (int i = 0; i < bucketNumber; i++)
           points = buckets[i].transfer(points);

       delete[] buckets;
}


void bsortPtrY (const Point ** points, const int n) { // bucket-sorts pointer array points by y
    const static int nEl = 5;

    class Bucket {
           int size = int(nEl * 2.5); // Heurstic
           const Point ** b = new const Point*[size]; // Array of pointers to Points
           int n = 0;
        
       public:
           void insert (const Point * p) {
               if (n >= size) {
                   size = size + nEl;
                   const Point ** bNew = new const Point*[size]; // Array of Point addresses
                   for (int i = 0; i < n; i++)
                       bNew[i] = b[i];
                   delete [] b;
                   b = bNew;
               }
               b[n++] = p;
            }
            
            void sort () { qsort(b, n, sizeof(Point*), comparePtrPointsY); }
            
            const Point ** transfer (const Point ** array) const {
                for (int i = 0; i < n; i++)
                    *(array++) = b[i];
                return array; // next to be transferred
            }
        
            ~Bucket () { delete [] b; }
        };
        
        int bucketNumber = n / nEl + 1;
        Bucket * buckets = new Bucket[bucketNumber]; // Array of buckets of Point pointers

        for (int i = 0; i < n; i++)  //Filling Buckets
            buckets[int((points[i]->getY() + 0.5) * bucketNumber)].insert(points[i]);

        for (int i = 0; i < bucketNumber; i++)
            buckets[i].sort();
        
        for (int i = 0; i < bucketNumber; i++)
            points = buckets[i].transfer(points);

        delete[] buckets;
}

double maxMinDist (int n) { // Maximal minimum distance between n points in the unit square
    double sqrt3 = sqrt(3);
    double c = (3 <= n && n <= 6) ? n - 2
            : (7 <= n && n <= 9) ? n - 1
            : 3 * n/2 - 5 + n % 2;
    double u1 = 2 / (sqrt(n*M_PI + c*(sqrt3 - M_PI/2) + (4*(int)sqrt(n) - 2)*(2-M_PI/2)) - 2);
    double u2 = (1 + sqrt(1 + (n-1)*2/sqrt3)) / (n-1);
    return fmin(u1, u2);
}

double maxMinDistRect (int n, double width, double height) { // maxMinDist for rectangle width x height
    double mmD1 = maxMinDist(n);
    double f = width * height;
    return mmD1 * sqrt(f / 1);
}

class Solver {
    const std::string id;
    
protected:
    Log* log;

    const Point* points;
    int n;
    double width, height;
    double minX, maxX, minY, maxY;
    double minDist;

    Solver (std::string id, const Point points[], int n) : points(points), n(n), id(id), log(Log::make(id)) {
        minX = points[0].getX();
        maxX = minX;
        minY = points[0].getY();
        maxY = minY;
        for (int i = 1; i < n; i++) { // O(n)
            double x = points[i].getX();
            double y = points[i].getY();
            if (x < minX)
                minX = x;
            else if (x > maxX)
                maxX = x;
            if (y < minY)
                minY = y;
            else if (y > maxY)
                maxY = y;
        }
        width = maxX - minX;
        height = maxY - minY;

        minDist = dist(Point(minX, minY), Point (maxX, maxY));
    }

public:
    virtual PointPair solve () = 0;

    virtual double getMinDist () const { return minDist; }
    
    virtual double dist (const Point& p1, const Point& p2) {
        double dx = p1.getX() - p2.getX();
        double dy = p1.getY() - p2.getY();
        return sqrt(dx*dx + dy*dy);
    }
    
    virtual ~Solver () {}

    std::string getId () { return id; }

protected:
    void solve3Points (int first, int last, PointPair& cpp) { // brute force minimum search for <= 3 points (points[first:last])
        switch (last - first + 1) { // n
            case 0:
            case 1:
                minDist = 2; // "infinity" in the unit square
                break;

            case 2:
                cpp.p1 = &points[first];
                cpp.p2 = &points[first + 1];
                minDist = dist(*cpp.p1, *cpp.p2);
                break;
            case 3: {
                    double d01 = dist(points[first],points[first + 1]);
                    double d02 = dist(points[first],points[first + 2]);
                    double d12 = dist(points[first + 1],points[first + 2]);
                    minDist = fmin(d01, fmin(d02, d12));
                    if (d01 == minDist) {
                        cpp.p1 = &points[first];
                        cpp.p2 = &points[first + 1];
                    } else if (d02 == minDist) {
                        cpp.p1 = &points[first];
                        cpp.p2 = &points[first + 2];
                    } else { // d12 == min
                        cpp.p1 = &points[first + 1];
                        cpp.p2 = &points[first + 2];
                    }
                }
                break;
            default:
                std::cout << "solve3Points: <0 elements!" << std::endl;
                exit(1);
        }
    }
    /*
    void updateMinDist (const Point & p1, const Point & p2, PointPair & cpp) {
        double d = dist(p1, p2);
        if (d < minDist) {
            minDist = d;
            cpp.p1 = &p1;
            cpp.p2 = &p2;
        }
    }
     */
};

// Macro version to compare the distance of two points with minDist(inline method didn't seem to work in-line):
#define updateMinDist(pt1, pt2, cpp) {  \
    double d = dist((pt1), (pt2));      \
    if (d < minDist) {                  \
        minDist = d;                    \
        (cpp).p1 = &(pt1);              \
        (cpp).p2 = &(pt2);              \
    }                                   \
}

class APSolver: public Solver {
public:
    APSolver (const Point points[], int n, std::string id = "AP") : Solver(id, points, n) {}

    virtual PointPair solve () {
        log->enter();
        PointPair cpp;
        for (int i1 = 0; i1 < n-1; i1++)
            for (int i2 = i1+1; i2 < n; i2++)
                updateMinDist(points[i1], points[i2], cpp);
        log->exit();
        return cpp;
    }
};

class APSSolver: public Solver {
public:
    APSSolver (const Point points[], int n, std::string id = "APs") : Solver (id, points, n) {}

    virtual PointPair solve () {
        Point* points = clone(this->points, n); // To avoide side effects in the point cloud

        log->enter();
        minDist = maxMinDistRect(n, width, height);
//        qsort(points, n, sizeof(Point), comparePointsX);
        bsort(points, n);

        PointPair cpp;
        for (int i1 = 0; i1 < n-1; i1++) {
            double x1 = points[i1].getX();
            for (int i2 = i1 + 1; i2 < n; i2++) {
                double dx = points[i2].getX() - x1;
                if (dx > minDist)
                    break; // ditance to point 1 cannot become < minDist
                updateMinDist(points[i1], points[i2], cpp);
            }
        }
        log->exit();
        delete [] points;
        return cpp;
    }
};

class KMSolver: public Solver {
public:
    KMSolver (const Point points[], int n, std::string id = "KM") : Solver(id, points, n) {}

    virtual PointPair solve () {
        log->enter();
        PointPair cpp;
        
        const Point** Si = new const Point*[n];
        for (int p = 0; p < n; p++)
            Si[p] = &points[p];
        int nSi = n;    // |Si|
        int i = 0;      // Iteration, will be incremented befor the loop body
        double dxi = 2; // Infinity, since we operate in the unit square
        
        while (nSi > 0) {
            i++;
            int xi;

			for (xi = 0; xi < n && Si[xi] == NULL; xi++); // Pick random point. Since our points are all random, this is ok for our tests

			dxi = 2;     // Infinity, since we operate in the unit square
            for (int p=0; p < n; p++) { // Loop is designed for "real" random points
                if (p != xi && Si[p] != NULL) {
                    double d = dist(*Si[xi], *Si[p]);
                    if (d < dxi)
                        dxi = d;
                }
            }
            double b = dxi / 3;
            int g = (int)(1/b)+1;
            Grid G(g, nSi); // Mesh of O(nSi) space
            for (int p = 0; p < n; p++)
                if (Si[p] != NULL) { // p in Si
                    int gi = (int)((Si[p]->getX() - minX) / b);
                    int gj = (int)((Si[p]->getY() - minY) / b);
                    G.insert(gi, gj, *Si[p]);
                }

            for (int p = 0; p < n; p++)
                if (Si[p] != NULL) { // p in Si
                    int gi = (int)((Si[p]->getX() - minX) / b);
                    int gj = (int)((Si[p]->getY() - minY) / b);
                    PointList* bucket = G.get(gi, gj);
                    if (bucket->next == NULL && // Alone at home, check neighbours
                        G.get(gi-1, gj-1) == NULL && G.get(gi-1, gj) == NULL && G.get(gi-1, gj+1) == NULL &&
                        G.get(gi, gj-1) == NULL && G.get(gi, gj+1) == NULL &&
                        G.get(gi+1, gj-1) == NULL && G.get(gi+1, gj) == NULL && G.get(gi+1, gj+1) == NULL) {
                        Si[p] = NULL; // Delete from Set
                        nSi--;
                    }
                }
        }
        delete [] Si;
        
        // Step 3, else: Si is empty
        // i* -> i, d(xi*) -> dxi -> minDist
        // Step 4
        int g = (int)(1/dxi)+1;
        Grid G(g, n); // Mesh of O(n) space
        for (int p = 0; p < n; p++) {
                int gi = (int)((points[p].getX() - minX) / dxi);
                int gj = (int)((points[p].getY() - minY) / dxi);
                G.insert(gi, gj, points[p]);
        }

		minDist = dxi;
        for (int p = 0; p < n; p++) {
            const Point& x = points[p];
            int gi = (int)((x.getX() - minX) / dxi);
            int gj = (int)((x.getY() - minY) / dxi);
            
            for (const PointList* l2 = G.get(gi, gj); l2 != NULL; l2 = l2->next)
                if (l2->p != &x)
                    updateMinDist(x, *l2->p, cpp);
            for (const PointList* l2 = G.get(gi-1, gj-1); l2 != NULL; l2 = l2->next)
                updateMinDist(x, *l2->p, cpp);
            for (const PointList* l2 = G.get(gi-1, gj); l2 != NULL; l2 = l2->next)
                updateMinDist(x, *l2->p, cpp);
            for (const PointList* l2 = G.get(gi-1, gj+1); l2 != NULL; l2 = l2->next)
                updateMinDist(x, *l2->p, cpp);
            for (const PointList* l2 = G.get(gi, gj-1); l2 != NULL; l2 = l2->next)
                updateMinDist(x, *l2->p, cpp);
            for (const PointList* l2 = G.get(gi, gj+1); l2 != NULL; l2 = l2->next)
                updateMinDist(x, *l2->p, cpp);
            for (const PointList* l2 = G.get(gi+1, gj-1); l2 != NULL; l2 = l2->next)
                updateMinDist(x, *l2->p, cpp);
            for (const PointList* l2 = G.get(gi+1, gj); l2 != NULL; l2 = l2->next)
                updateMinDist(x, *l2->p, cpp);
            for (const PointList* l2 = G.get(gi+1, gj+1); l2 != NULL; l2 = l2->next)
                updateMinDist(x, *l2->p, cpp);
        }
        log->exit();
        return cpp;
    }
};

class MMSolver: public Solver {

public:
    MMSolver (const Point points[], int n, std::string id="MM") : Solver(id, points, n) {}

    PointPair solve () {
        log->enter();
        minDist = maxMinDistRect(n, width, height);
        int gx = (int)(width/minDist)+1, gy = (int)(height/minDist)+1;
        
        PointList ***G = new PointList ** [gx];
        for (int i=0; i<gx; i++) {
            G[i] = new PointList*[gy];
            for (int j=0; j<gy; j++)
                G[i][j] = NULL;
        }

        for (int i = 0; i < n; i++) {
            int gi = (int)((points[i].getX() - minX) / minDist);
            int gj = (int)((points[i].getY() - minY) / minDist);
            G[gi][gj] = new PointList(points[i], G[gi][gj]);
        }
        
        PointPair cpp;
        for (int gi = 0; gi < gx; gi++) {
            for (int gj = 0; gj < gy; gj++)
                for (const PointList * l1 = G[gi][gj]; l1 != NULL; l1 = l1->next) {
                    const Point* p1 = l1->p;
                    for (const PointList * l2 = l1->next; l2 != NULL; l2 = l2->next)  // partners from same cell
                        updateMinDist(*p1, *l2->p, cpp);
                    if (gi < gx - 1)
                        for (const PointList * l2 = G[gi+1][gj]; l2 != NULL; l2 = l2->next) // partners from cell to the right
                            updateMinDist(*p1, *l2->p, cpp);
                    if (gj < gy - 1) {
                        for (const PointList * l2 = G[gi][gj+1]; l2 != NULL; l2 = l2->next) // partners from cell directly underneath
                            updateMinDist(*p1, *l2->p, cpp);
                        if (gi < gx - 1)
                            for (const PointList * l2 = G[gi+1][gj+1]; l2 != NULL; l2 = l2->next)  // partners from cell right underneath
                                updateMinDist(*p1, *l2->p, cpp);
                        if (gi > 0)
                            for (const PointList * l2 = G[gi-1][gj+1]; l2 != NULL; l2 = l2->next)  // partners from cell left underneath
                                updateMinDist(*p1, *l2->p, cpp);
                    }
                }
        }
        for (int gi=0; gi<gx; gi++) {
            for (int gj=0; gj<gy; gj++)
                // if (G[gi][gj] != NULL) // https://stackoverflow.com/questions/4190703/is-it-safe-to-delete-a-null-pointer
                    delete G[gi][gj];
            delete [] G[gi];
        }
        delete [] G;
        
        log->exit();
        return cpp;
    }
};

class RLSolverD: public Solver { // Rabin-Lipton with estimation of shortest distance as suggested by Lipton
public:
    RLSolverD (const Point p[], int n, std::string id = "RL") : Solver(id, p, n) {}

    PointPair solve () {
        log->enter();

        double delta = estimateMinDist(n-1);
        minDist = delta;

        int g = (int)(1/delta)+1;
        Grid G(g, n); // O(n) space
        for (int i = 0; i < n; i++) {
            int gi = (int)((points[i].getX() - minX) / delta);
            int gj = (int)((points[i].getY() - minY) / delta);
            G.insert(gi, gj, points[i]);
        }

        PointPair cpp;
        for (int i = 0; i < n; i++) {
            int gi = (int)((points[i].getX() - minX) / delta);
            int gj = (int)((points[i].getY() - minY) / delta);
            MarkedPointList* bucket = G.get(gi, gj);
            if (/* bucket != NULL && REDUNDANT */ !bucket->visited) {
                bucket->visited = true;
                for (const PointList* l1 = bucket; l1 != NULL; l1 = l1->next) {
                    const Point* p1 = l1->p;
                    for (const PointList* l2 = l1->next; l2 != NULL; l2 = l2->next)
                        updateMinDist(*p1, *l2->p, cpp);
                    if (gi < g - 1) {
                        for (const PointList* l2 = G.get(gi+1, gj); l2 != NULL; l2 = l2->next)
                            updateMinDist(*p1, *l2->p, cpp);
                        if (gj < g - 1)
                            for (const PointList* l2 = G.get(gi+1, gj+1); l2 != NULL; l2 = l2->next)
                                updateMinDist(*p1, *l2->p, cpp);
                        if (gj > 0)
                            for (const PointList* l2 = G.get(gi+1, gj-1); l2 != NULL; l2 = l2->next)
                                updateMinDist(*p1, *l2->p, cpp);
                    }
                    if (gj < g - 1)
                        for (const PointList* l2 = G.get(gi, gj+1); l2 != NULL; l2 = l2->next)
                            updateMinDist(*p1, *l2->p, cpp);
                }
            }
        }
        log->exit();
        return cpp;
    }

private:
    double estimateMinDist (int sampleSize) {
            assert(sampleSize < n); // Due to the simplified logic below: Take distances of first n point pairs
            double mD = (width+height); // "infinity"
            for (int i = 1; i <= sampleSize; i++) {
                double d = dist(points[i-1], points[i]);
                if (d < mD)
                    mD = d;
            }
            return mD;
        }
};


class PSSolver: public Solver { // Plainsweep Solver by Hinrichs et al.
    std::map<double, Point*> yTable; // <key = Y-coord., value = ptr. to Point>

    void insertY (Point& p) {
        yTable[p.getY()] = &p;
    }
    
    void deleteY (Point& p) {
        yTable.erase(p.getY());
    }
    
public:
    PSSolver (const Point points[], int n, std::string id = "PS") : Solver (id, points, n) {}

    virtual PointPair solve () {
        Point* xQueue = clone(this->points, n); // To avoide side effects in the point cloud

        log->enter();
        bsort(xQueue, n);
        
        PointPair cpp;
        cpp.p1 = & xQueue[0]; // closestLeft
        cpp.p2 = & xQueue[1]; // closestRight
        minDist = dist(xQueue[0], xQueue[1]);
        double & delta = minDist; // terminology adaptation
        insertY(xQueue[0]);
        insertY(xQueue[1]);
        
        int t = 0; // tail
        for (int c = 2; c < n; c++) {
            Point& current = xQueue[c];
            while (xQueue[t].getX() <= current.getX() - delta) {
                deleteY(xQueue[t]);
                t++;
            }
            
            for (auto check = yTable.upper_bound(current.getY() - delta);
                 check != yTable.end() && abs(check->first - current.getY()) <= delta;
                 check++) {
                if (check->second != &current) // Skip current (dist = 0)
                    updateMinDist(current, *(check->second), cpp); // = update delta
            }
            
            insertY(current);
        } // end transition
        
        log->exit();
        delete [] xQueue;
        return cpp;
    }
};


class DCSolver: public Solver {
public:
    DCSolver (const Point points[], int n, std::string id = "DC") : Solver(id, points, n) {}

    PointPair solve () {
        Point* pointsXSort = clone(this->points, n); // to avoid side effects
        const Point * pSafe = points;
        points = pointsXSort; // temporarily switch points to the sortable version

        log->enter();
        bsort(pointsXSort, n);

        const Point** pointsY = new const Point*[n];
        for (int i = 0; i < n; i++)
            pointsY[i] = &points[i];
        bsortPtrY(pointsY, n);
        PointPair cpp = cppDC(0, n-1, pointsY);

        log->exit();
        points = pSafe; // restore status quo ante
        delete [] pointsXSort;
        delete [] pointsY;
        return cpp;
    }

private:
    PointPair cppDC (int first, int last, const Point * pointsY[]) {
        log->enter();
        PointPair cpp;
        int n = last - first + 1;

        if (n <= 3)
            solve3Points(first, last, cpp);
        else {
            int mid = (first + last) / 2;
            double v = points[mid].getX(); // mid: rightmost point in P1 bzw. S1 or XL, resp.

            const Point ** yL = new const Point*[n/2 + n%2];
            const Point ** yR = new const Point*[n/2];
            int lYl = 0, lYr = 0; // Lengths of yL and yR
            for (int i = 0; i < n; i++)
                if (pointsY[i]->getX() <= v)
                    yL[lYl++] = pointsY[i];
                else
                    yR[lYr++] = pointsY[i];

            PointPair cpp1 = cppDC(first, mid, yL);
            double d1 = minDist;
            PointPair cpp2 = cppDC(mid + 1, last, yR);
            double d2 = minDist;
            if (d1 < d2) {
                minDist = d1;
                cpp = cpp1;
            } else {
                minDist = d2;
                cpp = cpp2;
            }

            const Point ** yPrime = new const Point*[lYr]; // Y' - all points in the d-strip, |Y'| <(=) n
            int lYp = 0;
            for (int i = 0; i < lYr; i++) {
                if (yR[i]->getX() <= v + minDist) // Point i within d
                    yPrime[lYp++] = yR[i];
            }

            if (lYp > 0) { // P2 nicht leer
                for (int i1 = mid; i1 >= first; i1--) { // Go eftwards in P1
                    const Point * p1 = &points[i1];
                    double x1 = p1->getX();
                    if (v - x1 > minDist)
                        break; // reached left margin of P1
                    double y1 = p1->getY();
                    for (int j = searchHighestPointLowerEq(y1 + minDist, yPrime, lYp-1); j >= 0; j--) { // Go down; j >= 0 only as emergency exit
                        const Point * p2 = yPrime[j];
                        double y2 = p2->getY();
                        if (x1 != p2->getX() || y1 != y2) { // p1 != p2
                            if (y1 - y2 > minDist)
                                break; // Reached lower margin of R
                            updateMinDist(*p1, *p2, cpp);
                        }
                    }
                }
            }
            delete [] yPrime;
            delete [] yL;
            delete [] yR;
        }
        log->exit();
        return cpp;
    }

    int searchHighestPointLowerEq (double y0, const Point * ySort[], int u) {
        int l = 0;

        while (u > l) { // > 1 element
            int m = (l + u) / 2;
            double y = ySort[m]->getY();
            if (y > y0)  // too high
                u = m - 1;
            else if (y == y0) // found
                return m;
            else // too low
                l = m + 1;
        }
        return l;
    }
};

bool compareDouble(double a, double b, double epsilon)
{
    return fabs(a - b) < epsilon;
}

struct PointPairDist {
    PointPair pair;
    double dist = std::numeric_limits<double>::max();
};

void testSeries (const int nMin, const int nMax, const double cloudwidth, const double cloudheight, const int tries) {
    
    srand(23021988 - 4101959 + 3011992); // srand((unsigned)time(NULL));
    
#ifndef SIGMA
    PointGenerator* randomGen[] = {
        new UDGenerator(-cloudwidth/2, cloudwidth/2, -cloudheight/2, cloudheight/2),
        new NDGenerator(-cloudwidth/2, cloudwidth/2, -cloudheight/2, cloudheight/2, 0.2)
    };
#else
    PointGenerator* randomGen[7];
    double sigma = 1;
    for (int g = 0; g<DIM(randomGen); g++, sigma /= 2)
        randomGen[g] = new NDGenerator(-cloudwidth/2, cloudwidth/2, -cloudheight/2, cloudheight/2, sigma);
#endif

    for (int d = 0; d<DIM(randomGen); d++) { // for all distributions d
    std::cout << "nMin: " << nMin << " nMax: " << nMax << " width: " << cloudwidth << " height: " << cloudheight << " tries: " << tries << " Generator: " << randomGen[d]->id() << std::endl;
    
        for (int n = nMin; n <= nMax; n *= 2) {
			Log::startAll();
            for (int t = 1; t <= tries; t++) {
                const Point* points = makePoints(n, *randomGen[d]);
                PointPair cpp;
                Solver* s[] = {
                    new APSSolver(points, n), // #0
                    new APSolver(points, n),  // #1
                    new DCSolver(points, n),
                    new PSSolver(points, n),
                    new KMSolver(points, n),
                    new RLSolverD(points, n),
                    new MMSolver(points, n),
                };
                int nSolvers = DIM(s);

                cpp = s[0]->solve();
                double dOpt = s[0]->getMinDist();
                
                for (int i = n <= 131072 ? 1 : 2; // Avoid APSolver for greater n
                                        i<nSolvers; i++) {
                    cpp = s[i]->solve();
                    if (s[i]->getMinDist() != dOpt) {
                        std::cout << s[i]->getId() << ": " << s[i]->getMinDist() << " > " << dOpt << std::endl;
                    }
                }

                delete [] points;
                for (int i=0; i<nSolvers; i++)
                    delete s[i];
            }
            Log::printAll(n, tries);
        }
        std::cout << std::endl;
    }
}


int main (int argc, const char * argv[]) {
#ifndef SIGMA
    int par[] = {1024, 33554432, 1, 1, 10}; // Semantics see below
#else
	int par[] = {16777216, 16777216, 1, 1, 10}; // Semantics see below
#endif
	testSeries(par[0], // nMin
			   par[1], // nMax
			   par[2], // cloudwidth
			   par[3], // cloudheight
			   par[4]);// tries
    
    std::cout << argv[0] << " done." << std::endl;
    return 0;
}
